"""
Start up a mitmweb instance using the authentication addons
"""

import os
from pathlib import Path
import shlex
import sys

from kerberos_auth_proxy.utils import env_to_options
from kerberos_auth_proxy.utils import dotenv_from_args


def setup_certificates():
    print("INFO: writing mitmproxy-ca.pem", file=sys.stderr)

    confdir = Path(os.environ["MITM_SET_CONFDIR"])
    crt = Path(os.environ["MITM_TLS_CA_CRT"])
    key = Path(os.environ["MITM_TLS_CA_KEY"])

    confdir.mkdir(parents=True, exist_ok=True)

    for old in confdir.glob("mitmproxy-ca*"):
        old.unlink()

    dest = confdir / "mitmproxy-ca.pem"
    dest.touch()
    dest.chmod(0o600)

    with dest.open("w") as bundle:
        bundle.write(crt.read_text())
        bundle.write(key.read_text())


def main():
    dotenv_from_args(sys.argv)
    setup_certificates()

    from kerberos_auth_proxy.mitm.addons import setup

    plugin_path = os.path.realpath(setup.__file__)
    env_options = list(env_to_options(os.environ))

    if len(sys.argv) <= 1 or sys.argv[1] not in ('mitmdump', 'mitmweb', 'mitmproxy'):
        raise Exception('No valid command')

    args = [sys.argv[1], "-s", plugin_path] + env_options + sys.argv[2:]
    print(f"$ {shlex.join(args)}", file=sys.stderr)

    return os.execvpe(args[0], args, os.environ)


if __name__ == "__main__":
    main()



import asyncio
from logging import getLogger
from typing import Optional
from urllib.parse import urlparse

import gssapi
from requests_gssapi import HTTPSPNEGOAuth
from mitmproxy.http import HTTPFlow

from kerberos_auth_proxy.kerberos import KerberosCache
from kerberos_auth_proxy.mitm.hostutils import UrlPattern
from kerberos_auth_proxy.mitm.username import get_username

LOGGER = getLogger(__name__)


class SpnegoAddon:
    def __init__(
        self,
        url_patterns: list[UrlPattern],
        default_username: Optional[str] = None,
    ) -> None:
        self.url_patterns = url_patterns
        self.default_username = default_username or None
        self.cache = KerberosCache()

    @classmethod
    def create(
        cls,
        urls: Optional[list[str]] = None, 
        default_username: Optional[str] = None,
    ) -> 'SpnegoAddon':
        return cls(
            url_patterns=[UrlPattern.create(url) for url in (urls or [])],
            default_username=default_username or None,
        )

    async def request(self, flow: HTTPFlow):
        if not (username := get_username(flow, self.default_username)):
            LOGGER.debug('skipping SPNEGO flow, no valid username')
            return

        flow_url = urlparse(flow.request.url)
        if not UrlPattern.matches_any(self.url_patterns, flow_url):
            LOGGER.debug('skipping SPNEGO flow, URL does not match: %s', flow.request.url)
            return

        negotiate = await generate_spnego_negotiate(
            cache=self.cache,
            authority=flow.request.host,
            username=username,
        )
        flow.request.headers[b'Authorization'] = negotiate


async def generate_spnego_negotiate(cache: KerberosCache, username: str, authority: str) -> str:
    host = authority.split(':')[0]
    principal = await cache.login(username)

    def _generate_spnego_negotiate_blocking(host: str, principal: str) -> str:
        LOGGER.info('generating Negotiate token for host=%s and principal=%s', host, principal)
        name = gssapi.Name(principal, gssapi.NameType.kerberos_principal)
        creds = gssapi.Credentials(name=name, usage="initiate")

        gssapi_auth = HTTPSPNEGOAuth(
            creds=creds,
            opportunistic_auth=True,
            target_name="HTTP",
        )
        return gssapi_auth.generate_request_header(None, host, True)

    loop = asyncio.get_event_loop()
    return await loop.run_in_executor(
        None, _generate_spnego_negotiate_blocking, host, principal
    )



import importlib
import os

from ruamel.yaml import YAML

conf_file = os.getenv('CONFIG_FILE') or 'config.yaml'

yaml = YAML(typ='safe', pure=True)
with open(conf_file) as fp:
    conf = yaml.load(fp)

addons = []

for addon_conf in (conf.get('addons') or []):
    module_name, _, cls_name = addon_conf['type'].partition(':')
    module = importlib.import_module(module_name)
    cls = getattr(module, cls_name)

    kwargs = addon_conf.get('kwargs') or {}
    instance = cls.create(**kwargs)

    addons.append(instance)



"""
Miscellaneous utilities for remapping hosts and URLs
"""

from dataclasses import dataclass
import re
from typing import Optional, Pattern
from urllib.parse import urlparse, urlunparse, ParseResult

from mitmproxy.http import Request, Response

REDIRECT_CODES = [301, 302, 303, 307, 307, 308]


def request_rebase(
    request: Request, matcher_url: ParseResult, target_url: ParseResult
) -> Optional[bool]:
    """
    Checks if the request url matches the matcher parameter and if so rebases it into the target URL

    >>> matcher_url = urlparse('http://example.external:8080/api/v1')
    >>> target_url = urlparse('http://example.internal:8080/api/v2')
    >>> request = Request.make('GET', 'http://example.external:8080/api/v1/do/stuff')
    >>> rebased = request_rebase(request, matcher_url, target_url)
    >>> rebased and request.path == '/api/v2/do/stuff'
    True

    >>> matcher_url = urlparse('http://example.com/')
    >>> target_url = urlparse('http://google.com/')
    >>> request = Request.make('GET', 'http://other')
    >>> request_rebase(request, matcher_url, target_url) is None
    True
    """
    if not url_parts_matches(matcher_url, request.host, request.port, request.path):
        return

    request.scheme = target_url.scheme
    request.host = target_url.hostname
    request.host_header = target_url.netloc
    request.port = url_port(target_url)
    request.path = path_rebase(matcher_url.path, request.path, target_url.path)

    return request


def redirect_rebase(
    response: Response, matcher_url: ParseResult, target_url: ParseResult
) -> bool:
    """
    Checks if the response is a redirect and if its Location matches the matcher parameter. If so,
    rebase it into the target URL

    >>> response = Response.make(200, headers={})
    >>> redirect_rebase(response, None, None)
    False

    >>> matcher_url = urlparse('http://internal:80/v1')
    >>> response = Response.make(301, headers={'Location': ''})
    >>> redirect_rebase(response, matcher_url, None)
    False

    >>> matcher_url = urlparse('http://internal:80/v1')
    >>> target_url = urlparse('http://external:80/v2')
    >>> response = Response.make(307, headers={'Location': 'http://non-matching/v1/resource'})
    >>> redirect_rebase(response, matcher_url, target_url)
    False

    >>> matcher_url = urlparse('http://internal:80/v1')
    >>> target_url = urlparse('http://external:80/v2')
    >>> response = Response.make(301, headers={'Location': 'http://internal/v1/resource'})
    >>> redirected = redirect_rebase(response, matcher_url, target_url)
    >>> redirected and response.headers[b'Location']
    'http://external/v2/resource'

    """
    if response.status_code not in REDIRECT_CODES:
        return False

    try:
        location = response.headers.get(b"Location")
        location_url = urlparse(location or "")
    except Exception:
        return False

    if not location_url.scheme:
        return False

    rebased_url = url_rebase(matcher_url, location_url, target_url)
    if rebased_url:
        response.headers[b"Location"] = urlunparse(rebased_url)
        return True

    return False


def url_rebase(
    matcher_url: ParseResult, url: ParseResult, target_url: ParseResult
) -> Optional[ParseResult]:
    """
    Checks if the url matches the matcher parameter and if so rebases it into the target URL

    >>> matcher_url = urlparse('http://example.external:8080/api/v1')
    >>> url = urlparse('http://example.external:8080/api/v1/path')
    >>> target_url = urlparse('http://internal:8081/v2')
    >>> url_rebase(matcher_url, url, target_url)
    ParseResult(scheme='http', netloc='internal:8081', path='/v2/path', params='', query='', fragment='')

    >>> matcher_url = urlparse('http://example.external:8080/api/v1')
    >>> url = urlparse('http://example.external:8080/api/v3/path')
    >>> target_url = urlparse('http://internal:8081/v2')
    >>> url_rebase(matcher_url, url, target_url)

    >>> matcher_url = ParseResult(scheme='http', netloc='internal:80', path='/v1', params='', query='', fragment='')
    >>> url = ParseResult(scheme='http', netloc='internal', path='/v1/resource', params='', query='', fragment='')
    >>> target_url = ParseResult(scheme='http', netloc='external:80', path='/v2', params='', query='', fragment='')
    >>> url_rebase(matcher_url, url, target_url)
    ParseResult(scheme='http', netloc='external', path='/v2/resource', params='', query='', fragment='')
    """
    if not url_parts_matches(matcher_url, url.hostname, url_port(url), url.path):
        return

    return url._replace(
        scheme=target_url.scheme,
        netloc=f"{url_netloc(target_url)}",
        path=path_rebase(matcher_url.path, url.path, target_url.path),
    )


def path_rebase(matcher_path: str, path: str, target_path: str) -> str:
    """
    Rebase a given path onto a target path by replacing the matched portion of the path

    >>> path_rebase('/v1/', '/v1/some/stuff', '/v2/')
    '/v2/some/stuff'
    """
    base_path = path[len(matcher_path):]
    return target_path + base_path


def url_parts_matches(matcher_url: ParseResult, hostname: str, port: int, path: str):
    """
    Checks if the hostname, port and path all match the matcher_url parameter

    >>> matcher_url = urlparse('http://localhost:8080/v1')
    >>> url_parts_matches(matcher_url, 'localhost', 8080, '/v1/sub/path')
    True

    >>> matcher_url = urlparse('http://localhost:8080/v1')
    >>> url_parts_matches(matcher_url, 'localhost', 8080, '/v2/sub/path')
    False

    >>> matcher_url = urlparse('http://localhost:8080/v1')
    >>> url_parts_matches(matcher_url, 'test.localhost', 8080, '/v1')
    False

    >>> matcher_url = urlparse('http://localhost:8080/v1')
    >>> url_parts_matches(matcher_url, 'localhost', 8081, '/v1')
    False
    """
    if port != url_port(matcher_url):
        return False

    if not path.startswith(matcher_url.path):
        return False

    if hostname != matcher_url.hostname:
        return False

    return True


def url_matches(matcher_url: str, url: str) -> bool:
    """
    Checks if the hostname, port and path all match the matcher_url parameter

    >>> url_matches('http://example/some', 'http://example/some/stuff')
    True

    >>> url_matches('http://example/some', 'http://example:80/some/stuff')
    True
    """
    matcher_url = urlparse(matcher_url)
    url = urlparse(url)
    return url_parts_matches(matcher_url, url.hostname, url_port(url), url.path)


def url_port(url: ParseResult) -> int:
    """
    Gets the url port or a default one based on the protocol.

    Raises:
        ValueError: no port set and unrecognized protocol

    >>> url = urlparse('http://localhost:8080/v1')
    >>> url_port(url)
    8080

    >>> url = urlparse('http://example.com/v1')
    >>> url_port(url)
    80

    >>> url = urlparse('https://example.com')
    >>> url_port(url)
    443

    >>> url = urlparse('unknown://example.com')
    >>> url_port(url)
    Traceback (most recent call last):
        ...
    ValueError: unknown scheme: 'unknown'
    """
    if url.port is not None:
        return url.port
    if url.scheme == "http":
        return 80
    elif url.scheme == "https":
        return 443
    else:
        raise ValueError(f"unknown scheme: '{url.scheme}'")


def url_netloc(url: ParseResult) -> str:
    """
    Gets the url netloc, removing the port if it's the default one based on the protocol

    >>> url = urlparse('http://localhost:8080/v1')
    >>> url_netloc(url)
    'localhost:8080'

    >>> url = urlparse('http://example.com/v1')
    >>> url_netloc(url)
    'example.com'

    >>> url = urlparse('https://example.com:443')
    >>> url_netloc(url)
    'example.com'
    """
    if url.scheme == "http" and url.port == 80:
        return url.hostname
    elif url.scheme == "https" and url.port == 443:
        return url.hostname
    else:
        return url.netloc


def url_default_port(scheme: str) -> int:
    """
    Gets the default url port for the given scheme

    Raises:
        ValueError: unrecognized protocol

    >>> url_default_port('http')
    80

    >>> url_default_port('https')
    443

    >>> url_default_port('unknown')
    Traceback (most recent call last):
        ...
    ValueError: unknown scheme 'unknown'
    """
    if scheme == "http":
        return 80
    elif scheme == "https":
        return 443
    else:
        raise ValueError(f"unknown scheme '{scheme}'")


@dataclass
class UrlPattern:
    scheme: str
    hostname: Pattern
    port: int
    path: Pattern

    @classmethod
    def create(cls, url: str) -> "UrlPattern":
        """
        Create a URL pattern from a URL string spec

        >>> pattern = UrlPattern.create('http://localhost:8080')
        >>> pattern
        UrlPattern(scheme='http', hostname=re.compile('(^|.)localhost$'), port=8080, path=re.compile('^/'))
        """
        parsed = urlparse(url)

        hostname = (parsed.hostname or "")
        hostname = hostname.replace("*", r"(.*?)")
        hostname = f"(^|.){hostname}$"

        port = parsed.port or url_default_port(parsed.scheme)

        path = (parsed.path) or "/"
        path = f"^{path}"

        return cls(
            scheme=parsed.scheme,
            hostname=re.compile(hostname),
            port=port,
            path=re.compile(path),
        )

    def match(self, url: ParseResult) -> bool:
        """
        Checks if the URL matches the pattern

        >>> pattern = UrlPattern.create('http://localhost')
        >>> [pattern.match(urlparse('http://localhost')),
        ...  pattern.match(urlparse('http://localhost:80/')),
        ...  pattern.match(urlparse('http://localhost/v1')),
        ...  pattern.match(urlparse('http://other.localhost')),
        ... ]
        [True, True, True, True]

        >>> pattern = UrlPattern.create('http://localhost')
        >>> [pattern.match(urlparse('https://extralocalhost')),
        ...  pattern.match(urlparse('http://localhost:8080/')),
        ... ]
        [False, False]
        """
        if url.scheme != self.scheme:
            return False

        if not self.hostname.search(url.hostname or ""):
            return False

        if (url.port or url_default_port(url.scheme)) != self.port:
            return False

        if not self.path.search(url.path or "/"):
            return False

        return True

    @classmethod
    def matches_any(cls, patterns: list["UrlPattern"], url: ParseResult) -> bool:
        """
        Checks if the URL matches any of the patterns

        >>> patterns = [UrlPattern.create('http://localhost'), UrlPattern.create('https://example.com:8443')]
        >>> UrlPattern.matches_any(patterns, urlparse('http://localhost'))
        True
        """
        return any(pattern.match(url) for pattern in patterns)



from pathlib import Path
import os
import secrets
import string
import subprocess
import sys
from typing import Optional

from kerberos_auth_proxy.utils import dotenv_from_args

DEFAULT_SUBJECT = '/O=Dev/OU=Development/CN=Kerberos Auth Proxy Self-Signed CA'


def generate_ca(ca_path: str, key_path: str, subject: Optional[str] = None):
    ca_path: Path = Path(ca_path)
    key_path: Path = Path(key_path)
    subject = subject or DEFAULT_SUBJECT

    if ca_path.exists():
        if not key_path.exists():
            raise Exception(f'CA {ca_path} without corresponding key {key_path}')
        elif not user_confirm(f'CA already exists at {ca_path}, recreate'):
            return

    ca_path.absolute().parent.mkdir(parents=True, exist_ok=True)
    key_path.absolute().parent.mkdir(parents=True, exist_ok=True)

    print(f'generating private key to {key_path}', file=sys.stderr)
    cmd = ['openssl', 'genrsa', '-out', str(key_path)]
    subprocess.run(cmd, check=True)

    print(f'generating self-signed CA to {ca_path} (subj={subject})')
    cmd = ['openssl', 'req', '-x509', '-new', '-nodes', '-key',
           str(key_path), '-sha256', '-days', '1825', '-out', str(ca_path), '-subj', subject]
    subprocess.run(cmd, check=True)


def generate_password(path: str) -> str:
    path: Path = Path(path)

    if path.exists():
        if not user_confirm(f'password file already exists at {path}, recreate'):
            return path.read_text()

    path.parent.mkdir(parents=True, exist_ok=True)
    path.touch()
    path.chmod(0o600)

    print(f'writing password to {path}', file=sys.stderr)
    chars = list(set(string.ascii_letters) - set('aeiouAEIOU') | set(string.digits))
    password = ''.join(secrets.choice(chars) for _ in range(15))

    path.write_text(password)
    return password


def generate_htpasswd(path: str, password: str, usernames: list[str]):
    path: Path = Path(path)

    if path.exists():
        if not user_confirm(f'.htpasswd file already exists at {path}, recreate'):
            return

    print(f'writing htpasswd file to {path}', file=sys.stderr)

    path.parent.mkdir(parents=True, exist_ok=True)
    path.touch()
    path.chmod(0o600)

    with path.open('w') as fp:
        for username in usernames:
            print(f'adding username {username!r} to htpasswd file', file=sys.stderr)
            cmd = ['htpasswd', '-n', '-i', '-B', username]
            process = subprocess.run(
                cmd,
                input=f'{password}\n',
                stdout=subprocess.PIPE,
                check=True,
                universal_newlines=True,
            )
            line = process.stdout.strip() + '\n'
            fp.write(line)


def user_confirm(prompt: str) -> bool:
    try:
        answer = input(f'{prompt} (y/N)? ')
    except EOFError:
        answer = 'N'

    answer = answer.strip().upper()
    if answer != 'Y':
        print(f'Cancelled.', file=sys.stderr)
        return False
    else:
        return True


def prepare(usernames: list[str]):
    ca_path = os.environ['MITM_TLS_CA_CRT']
    key_path = os.environ['MITM_TLS_CA_KEY']
    proxyauth = os.environ['MITM_OPT_PROXYAUTH']

    generate_ca(ca_path, key_path)

    if proxyauth.startswith('@'):
        htpasswd_path = proxyauth[1:]
        password_path = htpasswd_path + '.pwd'

        password = generate_password(password_path)
        generate_htpasswd(htpasswd_path, password, usernames)
    else:
        print(f'$MITM_OPT_PROXYAUTH does not point to an .htpasswd file, skipping', file=sys.stderr)


def main():
    dotenv_from_args(sys.argv)
    prepare(sys.argv[1:])


if __name__ == '__main__':
    main()



"""
Miscellaneous utilities
"""

from contextlib import contextmanager
from dotenv import load_dotenv
import os
import re
import sys
from pathlib import Path
import time
from typing import (
    Awaitable,
    Callable,
    Generator,
    Generic,
    Iterable,
    List,
    Mapping,
    Optional,
    TypeVar,
    Tuple,
)
import warnings

T = TypeVar("T")
Mapper = Callable[[str], T]


def string_to_map(s: Optional[str]) -> Mapping[str, str]:
    """
    Parses a string with a set of key=value items into a Python dict

    The values might be space or comma-delimited

    >>> string_to_map(None)
    {}

    >>> string_to_map('k1=v1 k2=v2')
    {'k1': 'v1', 'k2': 'v2'}

    >>> string_to_map('k1=v1,k2=v2')
    {'k1': 'v1', 'k2': 'v2'}

    >>> string_to_map('k1=v1 nokeyvalue ')
    Traceback (most recent call last):
        ...
    ValueError: invalid mapping 'nokeyvalue'
    """
    value = (s or "").replace(",", " ")
    parts = value.split()
    result = {}

    for part in parts:
        key, sep, value = part.partition("=")
        if not (key and sep):
            raise ValueError(f"invalid mapping {part!r}")
        result[key] = value

    return result


def string_to_list(s: Optional[str], mapper: Mapper) -> List[T]:
    """
    Explodes a string to a list of converted values

    The values might be space or comma-delimited

    >>> string_to_list('1 2', mapper=int)
    [1, 2]

    >>> string_to_list(None, mapper=int)
    []
    """
    s = s or ""
    return [mapper(item) for item in s.replace(",", " ").split() if item]


@contextmanager
def no_warnings(*categories) -> Generator[None, None, None]:
    """
    Disables the given warnings within the context
    """
    with warnings.catch_warnings():
        for category in categories:
            warnings.filterwarnings("ignore", category=category)
        yield


def _env_index(env_name: str) -> Tuple[str, int]:
    m = re.match(r".*_([0-9]+)$", env_name)
    if m:
        index = int(m.group(1))
        env_name = re.sub(r"_[0-9]+$", "", env_name)
        return [index, env_name]
    else:
        return [0, env_name]


def env_to_options(env: os._Environ) -> Iterable[str]:
    """
    Maps the environment variables to a set of mitm options

    >>> list(env_to_options({'MITM_SET_KERBEROS_REALM': 'LOCALHOST', 'MITM_SET_KERBEROS_SPNEGO_CODES': '401,407'}))
    ['--set', 'kerberos_realm=LOCALHOST', '--set', 'kerberos_spnego_codes=401,407']

    >>> list(env_to_options({'MITM_OPT_LISTEN_PORT': '3128'}))
    ['--listen-port', '3128']

    >>> list(env_to_options({'MITM_OPT_NO_WEB_OPEN_BROWSER': '-'}))
    ['--no-web-open-browser']

    >>> list(env_to_options({'MITM_OPT_MAP_REMOTE_1': 'v1', 'MITM_OPT_MAP_REMOTE_0': 'v0'}))
    ['--map-remote', 'v0', '--map-remote', 'v1']
    """

    # sort env alphabetically
    sorted_env = dict(sorted(env.items()), key=lambda item: item[0])
    # sort env by index suffix
    sorted_env = dict(sorted(env.items()), key=lambda item: _env_index(item[0])[1])

    for env_name, env_value in sorted_env.items():
        m = re.match(r".*_([0-9]+)$", env_name)
        if m:
            env_name = re.sub(r"_[0-9]+$", "", env_name)

        if env_name.startswith("MITM_SET_"):
            set_name = env_name[len("MITM_SET_") :].lower()
            yield "--set"
            yield f"{set_name}={env_value}"
        elif env_name.startswith("MITM_OPT_"):
            opt_name = env_name[len("MITM_OPT_") :].lower().replace("_", "-")
            yield f"--{opt_name}"
            if env_value != "-":
                yield env_value


class ExpiringCache(Generic[T]):
    def __init__(
        self,
        expiration: float,
        init: Callable[[str], Awaitable[T]],
    ):
        self.expiration = expiration
        self.init = init
        self._values: dict[str, T] = {}
        self._last_updates: dict[str, float] = {}

    async def get(self, key: str) -> Tuple[Optional[T], Optional[float]]:
        now = time.monotonic()
        last_update = self._last_updates.get(key)
        if last_update:
            age = now - last_update
        else:
            age = None

        if not age or age >= self.expiration:
            self._values[key] = await self.init(key)
            self._last_updates[key] = time.monotonic()
            age = None

        return self._values[key], age


def dotenv_from_args(argv: list[str]) -> Optional[Path]:
    """
    Recognizes a --env-file argument passed
    """
    print("INFO: checking --env-file arg")

    if len(argv) < 2:
        return

    if argv[1] == "--env-file":
        if len(argv) >= 3:
            env_path = argv[2]
            argv.pop(2)
            argv.pop(1)
        else:
            env_path = None
    elif argv[1].startswith("--env-file="):
        env_path = argv[1][len("--env-file=") :]
        argv.pop(1)
    else:
        return

    if not env_path:
        raise Exception("no value set for --env-file")

    env_path = Path(env_path).absolute()
    print(f"INFO: loading .env from {env_path}", file=sys.stderr)
    load_dotenv(dotenv_path=env_path, verbose=True, override=True)

    print(f"INFO: switching to directory {env_path.parent}")
    os.chdir(env_path.parent)

    return env_path



import asyncio
from logging import getLogger
import os
import time
from typing import Optional

from timelength import TimeLength

LOGGER = getLogger(__name__)


class KerberosCache:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._initialize()
        return cls._instance

    def _initialize(self):
        cache_name = os.getenv("KRB5CCNAME") or ""
        if not cache_name.startswith("DIR:"):
            raise ValueError("$KRB5CCNAME should be set to a DIR: type")

        self.realm = os.getenv('KERBEROS_REALM') or 'LOCALHOST'
        self.keytabs_path = os.getenv('KEYTABS_PATH') or '/etc/security/keytabs'
        self.expiration = TimeLength(os.getenv('CACHE_EXPIRATION') or '30m').total_seconds
        self.last_kinits: dict[str, float] = {}
        self.principals: dict[str, str] = {}
        self.lock = asyncio.Lock()

    async def get_principal(
        self, username: str, refresh: bool = False
    ) -> Optional[str]:
        if username in self.principals and not refresh:
            return self.principals[username]

        keytab_path = self.get_keytab_path(username)

        LOGGER.debug(f"getting principal from keytab {keytab_path}")
        principal = await self.get_principal_from_keytab(keytab_path)
        if not principal:
            LOGGER.info(f"no credencials available for user {username!r}")
            return

        self.principals[username] = principal
        return principal

    def get_keytab_path(self, username: str) -> str:
        return os.path.join(self.keytabs_path, username + ".keytab")

    async def login(self, username: str, refresh: bool = False) -> Optional[str]:
        """
        Returns a logged-in principal corresponding to the username

        None is returned in case of failure
        """
        principal = await self.get_principal(username, refresh=True)
        if not principal:
            LOGGER.info(f"no credencials available for user {username!r}")
            return

        keytab_path = self.get_keytab_path(username)

        LOGGER.debug(
            f"principal for {username!r} is {principal}, now acquiring cache lock"
        )

        async with self.lock:
            if not refresh and self.has_valid_login(username):
                LOGGER.info(f"now using cached credentials for user {username!r}")
                return principal

            LOGGER.info(
                f"getting credencials for {principal} from keytab {keytab_path}"
            )
            process = await asyncio.create_subprocess_exec(
                "kinit",
                "-kt",
                keytab_path,
                principal,
            )
            await process.communicate()
            if process.returncode != 0:
                LOGGER.warn(
                    f"failed to authenticate {username} using principal {principal}"
                )
                return

            self.last_kinits[username] = time.monotonic()
            self.principals[principal] = principal

            LOGGER.debug(
                f"successfully authenticated {principal} from keytab {keytab_path}"
            )
            return principal

    def has_valid_login(self, username) -> bool:
        last_kinit = self.last_kinits.get(username)
        return last_kinit and time.monotonic() - last_kinit <= self.expiration

    async def get_principal_from_keytab(self, keytab_path: str) -> Optional[str]:
        process = await asyncio.create_subprocess_exec(
            "klist",
            "-kt",
            keytab_path,
            stdout=asyncio.subprocess.PIPE,
        )
        stdout, _ = await process.communicate()
        if process.returncode != 0:
            return None

        stdout: str = stdout and stdout.decode("ascii") or ""
        for line in stdout.splitlines():
            parts = line.split(" ")
            if parts and parts[-1].endswith("@" + self.realm):
                return parts[-1]



from logging import getLogger
import re
from typing import Optional

from mitmproxy.http import HTTPFlow


LOGGER = getLogger(__name__)
USERNAME_PATTERN = re.compile(r'[a-z_][a-z0-9_]*')


def get_username(flow: HTTPFlow, default_username: Optional[str] = None) -> Optional[str]:
    proxy_auth = flow.metadata.get("proxyauth")
    if not proxy_auth or not proxy_auth[0]:
        if default_username:
            return default_username
        LOGGER.debug("no authenticated user via proxyauth")
        return

    username = proxy_auth[0]
    if not USERNAME_PATTERN.match(username):
        LOGGER.warning("invalid username in proxyauth: %r", username)
        return

    return username



from logging import getLogger
import os
import ssl
from typing import Optional
from urllib.parse import urlparse

import aiohttp
from mitmproxy.http import HTTPFlow
from timelength import TimeLength

from kerberos_auth_proxy.kerberos import KerberosCache
from kerberos_auth_proxy.mitm.hostutils import UrlPattern
from kerberos_auth_proxy.mitm.username import get_username

from kerberos_auth_proxy.mitm.addons.spnego import generate_spnego_negotiate
from kerberos_auth_proxy.utils import ExpiringCache

LOGGER = getLogger(__name__)


class BearerSpnegoAddon:
    def __init__(
        self,
        url_patterns: list[UrlPattern],
        token_url: str,
        token_method: Optional[str] = None,
        token_headers: Optional[dict[str, str]] = None,
        expiration: Optional[float] = None,
        default_username: Optional[str] = None,
    ) -> None:
        self.url_patterns = url_patterns
        self.token_url = token_url
        self.token_method = token_method or 'POST'
        self.token_headers = token_headers or {}
        self.expiration = expiration or 10*60
        self.kerberos_cache = KerberosCache()
        self.token_cache = ExpiringCache[str](
            init=self._get_token,
            expiration=expiration or 10*60,
        )
        self.default_username = default_username or None

    async def _get_token(self, username: str) -> str:
        return await generate_bearer_spnego_token(
            cache=self.kerberos_cache,
            username=username,
            token_url=self.token_url,
            token_method=self.token_method,
            token_headers=self.token_headers,
        )

    @classmethod
    def create(
        cls,
        urls: list[str],
        token_url: str,
        token_method: Optional[str] = None,
        token_headers: Optional[dict[str, str]] = None,
        expiration: Optional[str] = None,
        default_username: Optional[str] = None,
    ) -> 'BearerSpnegoAddon':
        return cls(
            url_patterns=[UrlPattern.create(url) for url in (urls or [])],
            token_url=token_url,
            token_method=token_method,
            token_headers=token_headers,
            expiration=TimeLength(expiration).total_seconds if expiration else None,
            default_username=default_username or None,
        )

    async def request(self, flow: HTTPFlow):
        if not (username := get_username(flow, self.default_username)):
            LOGGER.debug('skipping Bearer SPNEGO flow, no valid username')
            return

        flow_url = urlparse(flow.request.url)
        if not UrlPattern.matches_any(self.url_patterns, flow_url):
            LOGGER.debug('skipping Bearer SPNEGO flow, URL does not match: %s', flow.request.url)
            return

        token, _ = await self.token_cache.get(username)
        if token:
            flow.request.headers[b'Authorization'] = f'Bearer {token}'


async def generate_bearer_spnego_token(
    cache: KerberosCache,
    username: str,
    token_url: str,
    token_method: str,
    token_headers: Optional[dict[str, str]] = None,
) -> Optional[str]:
    token_headers = token_headers or {}
    authority = urlparse(token_url).netloc
    negotiate = await generate_spnego_negotiate(
        cache=cache,
        username=username,
        authority=authority,
    )
    headers = {**token_headers, 'Authorization': negotiate}

    LOGGER.info('generating Bearer SPNEGO token from %s %s', token_method, token_url)

    cafile = os.environ['MITM_SET_SSL_VERIFY_UPSTREAM_TRUSTED_CA']
    ciphers = os.getenv('SSL_CONTEXT_CIPHERS') or ''
    ssl_context = ssl.create_default_context(cafile=cafile)
    if ciphers:
        ssl_context.set_ciphers(ciphers)
    connector = aiohttp.TCPConnector(ssl=ssl_context)

    async with aiohttp.ClientSession(connector=connector) as session:
        async with session.request(method=token_method, url=token_url, headers=headers) as response:
            if response.ok:
                return await response.text()
            else:
                LOGGER.warning('request failed with status=%d: %r', response.status, await response.text())
